# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from tempfile import mkstemp
from ovos_utils.log import LOG
from neon_utils.packaging_utils import get_package_dependencies


def patch_config(config: dict = None):
    """
    Write the specified speech configuration to the global config file
    :param config: Mycroft-compatible configuration override
    """
    from ovos_config.config import LocalConf
    from ovos_config.locations import USER_CONFIG
    LOG.warning(f"Patching configuration with: {config}")
    config = config or dict()
    local_config = LocalConf(USER_CONFIG)
    local_config.update(config)
    local_config.store()


def _plugin_to_package(plugin: str) -> str:
    """
    Get a PyPI spec for a known plugin entrypoint
    :param plugin: plugin spec (i.e. config['tts']['module'])
    :returns: package name associated with `plugin` or `plugin`
    """
    known_plugins = {
        "neon-tts-plugin-glados": "neon-tts-plugin-glados",
        "neon_tts_mimic": "neon-tts-plugin-mimic",
        "amazon_polly": "neon-tts-plugin-polly",
        "coqui": "neon-tts-plugin-coqui",
        "neon-tts-plugin-larynx-server": "neon-tts-plugin-larynx-server",
        "mozilla_local": "neon-tts-plugin-mozilla-local",
        "mozilla_remote": "neon-tts-plugin-mozilla-remote"
    }
    return known_plugins.get(plugin) or plugin


def install_tts_plugin(plugin: str) -> bool:
    """
    Install a tts plugin using pip
    :param plugin: entrypoint of plugin to install
    :returns: True if the plugin installation is successful
    """
    import pip
    _, tmp_file = mkstemp()
    with open(tmp_file, 'w') as f:
        constraints = '\n'.join(get_package_dependencies("neon-audio"))
        f.write(constraints)
        LOG.info(f"Constraints={constraints}")
    LOG.info(f"Requested installation of plugin: {plugin}")
    returned = pip.main(['install', _plugin_to_package(plugin), "-c", tmp_file])
    LOG.info(f"pip status: {returned}")
    return returned == 0


def init_tts_plugin(plugin: str):
    """
    Initialize a specified plugin. Useful for doing one-time initialization
    before deployment
    """
    from ovos_plugin_manager.tts import load_tts_plugin
    plug = load_tts_plugin(plugin)
    if plug:
        LOG.info(f"Initializing plugin: {plugin}")
        plug()
    else:
        LOG.warning(f"Could not find plugin: {plugin}")


def use_neon_audio(func):
    """
    Wrapper to ensure call originates from neon_audio for stack checks.
    This is used for ovos-utils config platform detection which uses the stack
    to determine which module config to return.
    """
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest

from unittest.mock import Mock, patch
from threading import Event
from time import time
from ovos_bus_client import Message
from ovos_utils.messagebus import FakeBus
from neon_utils.message_utils import dig_for_message

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_audio.service import NeonPlaybackService


_TEST_CONFIG = {
    "g2p": {"module": "dummy"},
    "Audio": {},
    "tts": {"module": "neon-tts-plugin-larynx-server",
            "neon-tts-plugin-larynx-server": {"host": "https://larynx.2022.us/"}
            }
}


class TestAPIMethods(unittest.TestCase):
    bus = FakeBus()
    bus.connected_event = Event()
    bus.connected_event.set()
    test_config_dir = os.path.join(os.path.dirname(__file__), "config")
    os.environ["XDG_CONFIG_HOME"] = test_config_dir

    @classmethod
    @patch("ovos_audio.service.Configuration")
    def setUpClass(cls, config) -> None:
        config.return_value = _TEST_CONFIG
        ready_event = Event()

        def on_ready():
            ready_event.set()

        cls.audio_service = NeonPlaybackService(daemonic=True, bus=cls.bus,
                                                ready_hook=on_ready,
                                                disable_ocp=True)
        assert cls.audio_service.config == _TEST_CONFIG
        cls.audio_service.start()
        if not ready_event.wait(30):
            raise TimeoutError("Audio service not ready")

    @classmethod
    def tearDownClass(cls) -> None:
        try:
            cls.audio_service.shutdown()
        except Exception as e:
            print(e)

    def test_get_tts_no_sentence(self):
        context = {"client": "tester",
                   "ident": "123",
                   "user": "TestRunner"}
        tts_resp = self.bus.wait_for_response(Message("neon.get_tts", {},
                                                      dict(context)),
                                              context["ident"])
        for key in context:
            if key != "timing":
                self.assertEqual(context[key], tts_resp.context[key])
        self.assertIsInstance(tts_resp.context['timing']['response_sent'],
                              float, tts_resp.context['timing'])
        self.assertIsInstance(tts_resp.data.get("error"), str)
        self.assertEqual(tts_resp.data["error"], "No text provided.")

    def test_get_tts_invalid_type(self):
        context = {"client": "tester",
                   "ident": "1234",
                   "user": "TestRunner"}
        tts_resp = self.bus.wait_for_response(Message("neon.get_tts",
                                                      {"text": 123},
                                                      dict(context)),
                                              context["ident"], timeout=60)
        for key in context:
            self.assertEqual(context[key], tts_resp.context[key])
        self.assertIsInstance(tts_resp.context['timing']['response_sent'],
                              float, tts_resp.context['timing'])
        self.assertTrue(tts_resp.data.get("error")
                        .startswith("text is not a str:"))

    def test_get_tts_valid_default(self):
        text = "This is a test"
        context = {"client": "tester",
                   "ident": str(time()),
                   "user": "TestRunner"}
        tts_resp = self.bus.wait_for_response(Message("neon.get_tts",
                                                      {"text": text},
                                                      dict(context)),
                                              context["ident"], timeout=60)
        for key in context:
            self.assertEqual(context[key], tts_resp.context[key])
        self.assertIsInstance(tts_resp.context['timing']['response_sent'],
                              float, tts_resp.context['timing'])
        responses = tts_resp.data
        self.assertIsInstance(responses, dict)
        print(responses)
        self.assertEqual(len(responses), 1)
        resp = list(responses.values())[0]
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp.get("sentence"), text)

    # TODO: Test with multiple languages
    def test_get_tts_valid_speaker(self):
        pass

    def test_handle_speak(self):
        self.audio_service._playback_timeout = 1  # Override playback timeout
        real_method = self.audio_service.execute_tts
        mock_tts = Mock()
        self.audio_service.execute_tts = mock_tts

        session = {"session_id": "test_session"}

        # 'audio' in destination
        message_valid_destination = Message("speak",
                                            {"utterance": "test1"},
                                            {"ident": "test2",
                                             "destination": ['invalid',
                                                             'audio'],
                                             "session": session})
        self.audio_service.handle_speak(message_valid_destination)
        mock_tts.assert_called_with("test1", "test_session", False,
                                    message_valid_destination)

        # str 'audio' destination
        message_valid_destination = Message("speak",
                                            {"utterance": "test5"},
                                            {"ident": "test6",
                                             "destination": 'audio',
                                             "session": session})
        self.audio_service.handle_speak(message_valid_destination)
        mock_tts.assert_called_with("test5", "test_session", False,
                                    message_valid_destination)

        # Setup bus API handling
        self.audio_service._playback_timeout = 5
        msg = None

        def handle_tts(*args, **kwargs):
            nonlocal msg
            msg = dig_for_message()
            ident = msg.context.get('speak_ident')
            if ident:
                self.bus.emit(Message(ident))

        mock_tts.side_effect = handle_tts

        # Test No ident handling
        message_no_ident = Message("speak",
                                   {"utterance": "No Ident"},
                                   {"destination": ["audio"]})
        start_time = time()
        self.audio_service.handle_speak(message_no_ident)
        self.assertAlmostEqual(time(), start_time, 0)
        self.assertEqual(msg, message_no_ident)

        # Test `ident`
        ident = time()
        message_with_ident = Message("speak",
                                     {"utterance": "with ident",
                                      "ident": ident},
                                     {"destination": ["audio"]})
        on_ident = Mock()
        self.bus.on(ident, on_ident)
        self.audio_service.handle_speak(message_with_ident)
        self.assertEqual(msg, message_with_ident)
        on_ident.assert_called_once()

        # Test `speak_ident`
        speak_ident = time()
        message_with_speak_ident = Message("speak",
                                           {"utterance": "with speak ident",
                                            "ident": ident,
                                            "speak_ident": speak_ident},
                                           {"destination": ["audio"]})
        on_speak_ident = Mock()
        self.bus.on(speak_ident, on_speak_ident)
        self.audio_service.handle_speak(message_with_speak_ident)
        self.assertEqual(msg, message_with_speak_ident)
        on_ident.assert_called_once()
        on_speak_ident.assert_called_once()

        self.audio_service.execute_tts = real_method

    def test_get_tts_supported_languages(self):
        real_tts = self.audio_service.tts
        resp = self.bus.wait_for_response(Message(
            "ovos.languages.tts", {}, {'ctx': True}
        ))
        self.assertIsInstance(resp, Message)
        self.assertTrue(resp.context.get('ctx'))

        self.assertEqual(resp.data['langs'],
                         list(real_tts.available_languages) or ['en-us'])

        mock_languages = ('en-us', 'es', 'fr-fr', 'fr-ca')
        from ovos_plugin_manager.templates.tts import TTS

        class MockTTS(TTS):
            def __init__(self):
                super(MockTTS, self).__init__()

            @property
            def available_languages(self):
                return mock_languages

            def execute(self, *args, **kwargs):
                pass

        mock_tts = MockTTS()
        self.audio_service.tts = mock_tts
        resp = self.bus.wait_for_response(Message(
            "ovos.languages.tts", {}, {'ctx': True}
        ))
        self.assertEqual(resp.data['langs'], list(mock_languages))

        self.audio_service.tts = real_tts


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import shutil
import sys
import unittest

from time import time
from os.path import join, dirname
from threading import Event
from unittest.mock import Mock, patch
from click.testing import CliRunner
from ovos_bus_client import Message
# from ovos_plugin_manager.templates.tts import PlaybackThread
from ovos_utils.messagebus import FakeBus

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_audio.tts import WrappedTTS

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from test_objects import DummyTTS, DummyTTSValidator


class TTSBaseClassTests(unittest.TestCase):
    lang = "en-us"
    config = {"key": "val"}
    test_cache_dir = join(dirname(__file__), "test_cache")
    test_conf_dir = join(dirname(__file__), "config")
    bus = FakeBus()
    bus.connected_event = Event()
    bus.connected_event.set()

    @classmethod
    @patch("ovos_plugin_manager.templates.tts.Configuration")
    def setUpClass(cls, config) -> None:
        config.return_value = cls.config  # Explicitly no g2p
        os.makedirs(cls.test_conf_dir, exist_ok=True)
        os.environ["XDG_CACHE_HOME"] = cls.test_cache_dir

        cls.tts = WrappedTTS(DummyTTS, cls.lang, cls.config)
        cls.tts.init(cls.bus)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tts.shutdown()
        if os.path.exists(cls.test_cache_dir):
            shutil.rmtree(cls.test_cache_dir)
        if os.path.exists(cls.test_conf_dir):
            shutil.rmtree(cls.test_conf_dir)
        os.environ.pop("XDG_CACHE_HOME")

    def test_class_init(self):
        from ovos_plugin_manager.templates.tts import TTS
        # self.assertIsInstance(self.tts, WrappedTTS)
        self.assertIsInstance(self.tts, TTS)
        self.assertIsInstance(self.tts.bus, FakeBus)
        self.assertIsInstance(self.tts.language_config, dict)
        # TODO: Fix import errors in unit tests
        # self.assertIsNotNone(self.tts.lang_detector)
        # self.assertIsNotNone(self.tts.translator)
        self.assertEqual(self.tts.lang, self.lang)

        self.assertEqual(self.tts.config, self.config)
        self.assertIsInstance(self.tts.validator, DummyTTSValidator)
        self.assertFalse(self.tts.phonetic_spelling)
        self.assertEqual(self.tts.ssml_tags, ["speak"])

        self.assertEqual(self.tts.voice, "default")
        self.assertTrue(self.tts.queue.empty())

        from ovos_audio.playback import PlaybackThread
        from neon_audio.tts.neon import NeonPlaybackThread
        self.assertIsInstance(self.tts.playback, PlaybackThread)
        self.assertIsInstance(self.tts.playback, NeonPlaybackThread)

        self.assertIsInstance(self.tts.spellings, dict)
        self.assertEqual(self.tts.tts_name, "DummyTTS")
        self.assertIsInstance(self.tts.keys, dict)

        self.assertTrue(os.path.isdir(self.tts.cache_dir))
        # self.assertTrue(os.path.isfile(self.tts.translation_cache))
        self.assertIsInstance(self.tts.cached_translations, dict)

    def test_modify_tag(self):
        # TODO: Legacy
        self.assertEqual("test", self.tts.modify_tag("test"))

    def test_validate_ssml(self):
        # TODO: Legacy
        valid_tag_string = "<speak>hello</speak>"
        extra_tags_string = "<speak>hello</br></speak>"

        self.assertEqual(valid_tag_string, self.tts.validate_ssml(valid_tag_string))
        self.assertEqual(valid_tag_string, self.tts.validate_ssml(extra_tags_string))

    def test_preprocess_sentence(self):
        # TODO: Legacy
        sentence = "this is a test"
        self.assertEqual(self.tts._preprocess_sentence(sentence), [sentence])

    def test_execute(self):
        from neon_utils.signal_utils import check_for_signal, init_signal_bus
        init_signal_bus(self.bus)
        sentence = "testing"
        ident = time()
        default_execute = self.tts._execute
        self.tts._execute = Mock()
        self.tts.execute(sentence, ident)
        self.assertTrue(check_for_signal("isSpeaking"))
        self.tts._execute.assert_called_once_with(sentence, ident, False)
        self.tts._execute = default_execute

        default_get_multiple_tts = self.tts.get_multiple_tts
        self.tts.get_multiple_tts = Mock(return_value=dict())
        message = Message("test")
        self.tts.execute(sentence, ident, message=message)
        self.tts.get_multiple_tts.assert_called_once_with(message)

        # Test klat response
        klat_response = Mock()
        # Check called with klat_data context
        self.tts.bus.once('klat.response', klat_response)
        message.context['klat_data'] = dict()
        self.tts.execute(sentence, ident, message=message)
        klat_response.assert_called_once()
        # Check not called without klat_data context
        self.tts.bus.once('klat.response', klat_response)
        message.context.pop('klat_data')
        self.tts.execute(sentence, ident, message=message)
        klat_response.assert_called_once()

        self.tts.get_multiple_tts = default_get_multiple_tts

    def test_get_multiple_tts(self):
        # TODO
        pass

    def test_viseme(self):
        # TODO: Legacy
        self.assertIsNone(self.tts.viseme(""))

    def test_validator_valid(self):
        self.assertTrue(self.tts.validator.validate_lang())
        self.assertTrue(self.tts.validator.validate_dependencies())
        self.assertTrue(self.tts.validator.validate_connection())

    @patch("ovos_plugin_manager.templates.tts.Configuration")
    def test_validator_invalid(self, config):
        config.return_value = self.config  # Explicitly no g2p
        tts = DummyTTS("es", {})

        with self.assertRaises(Exception):
            tts.validator.validate()

        tts.shutdown()

    def test_get_tts(self):
        test_file_path = join(dirname(__file__), "test.wav")
        file, phonemes = self.tts._get_tts("test", wav_file=test_file_path,
                                           speaker={})
        self.assertEqual(file, test_file_path)
        self.assertIsNone(phonemes)


class TTSUtilTests(unittest.TestCase):
    def test_install_tts_plugin(self):
        from neon_audio.utils import install_tts_plugin
        self.assertTrue(install_tts_plugin("coqui"))
        self.assertTrue(install_tts_plugin("neon-tts-plugin-coqui"))
        self.assertFalse(install_tts_plugin("neon-tts-plugin-invalid"))

    def test_patch_config(self):
        import json
        from neon_audio.utils import use_neon_audio
        from neon_utils.configuration_utils import init_config_dir
        test_config_dir = os.path.join(os.path.dirname(__file__), "config")
        os.makedirs(test_config_dir, exist_ok=True)
        os.environ["XDG_CONFIG_HOME"] = test_config_dir
        use_neon_audio(init_config_dir)()

        with open(join(test_config_dir, "OpenVoiceOS", 'ovos.conf')) as f:
            ovos_conf = json.load(f)
        self.assertEqual(ovos_conf['submodule_mappings']['neon_audio'],
                         "neon_core")
        self.assertIsInstance(ovos_conf['module_overrides']['neon_core'], dict)

        from neon_audio.utils import patch_config
        import yaml
        test_config = {"new_key": {'val': True}}
        patch_config(test_config)
        conf_file = os.path.join(test_config_dir, 'neon',
                                 'neon.yaml')
        self.assertTrue(os.path.isfile(conf_file))
        with open(conf_file) as f:
            config = yaml.safe_load(f)

        self.assertTrue(config['new_key']['val'])
        shutil.rmtree(test_config_dir)
        os.environ.pop("XDG_CONFIG_HOME")


class TestCLI(unittest.TestCase):
    runner = CliRunner()

    @patch("neon_audio.cli.init_config_dir")
    @patch("neon_audio.__main__.main")
    def test_run(self, main, init_config):
        from neon_audio.cli import run
        self.runner.invoke(run)
        init_config.assert_called_once()
        main.assert_called_once()


if __name__ == '__main__':
    unittest.main()
